from sklearn.cluster import KMeans
import numpy as np

def get_optimal_cluster(
    points=None,
    max_k=20
):
    inertia_list = []
    for n_clusters in range(1, max_k+1):
        if len(points) >= n_clusters:
            model = KMeans(
                n_clusters=n_clusters,
                init="k-means++",
                max_iter=100,
                n_init=10,
                random_state=0
            ).fit(points)

            inertia_list.append(model.inertia_)

    k = [i*100 for i in np.diff(inertia_list, 2)].index(
        min(
            [i*100 for i in np.diff(inertia_list, 2)]
        )
    )

    return k



from fuzzywuzzy import fuzz
from datetime import datetime


def word_match_score(
    a,
    b
):
    score = fuzz.ratio(a, b)
    diff_len = abs(len(a)-len(b)) + 1
    diff_len = 1 if diff_len == 0 else diff_len
    return score//diff_len


def string_match(
    source=None,
    sample=None,
    threshold=99
):
    source_split = source.split()
    sample_split = sample.split()
    score = sum(
        word_match_score(a, b) for a, b in zip(source_split, sample_split)
    )
    score = score//max(len(source_split), len(sample_split))
    verified = (score >= threshold)
    return verified, score, threshold


def string_to_date(
    date_string
):
    date_time_obj = datetime.strptime(date_string, '%d.%m.%Y').date()

    return date_time_obj


if __name__ == "__main__":

    date_string = '14.08.1992'
    print(
        string_to_date(
            date_string
        )
    )



import cv2
from sabhi_utils.image_utils.utils import get_image, get_hsv_means_for_windows
from sabhi_utils.image_utils.utils import pad, load_image
from matplotlib import pyplot as plt
from scipy.stats import wasserstein_distance
import numpy as np
import imutils
import skimage.metrics as metrics
from skimage.io import imshow
from scipy import ndimage
import pandas as pd
import skimage.measure as measure


class GrayscaleHistogramComparison:

    def __init__(
        self,
        hist_size=256,
        hist_range=(0, 256),
        norm_type=cv2.NORM_MINMAX,
        comparison_method=wasserstein_distance,
        debug=False
    ):

        self._hist_size = hist_size
        self._hist_range = hist_range
        self._norm_type = norm_type
        self._comparison_method = comparison_method
        self._debug = debug

    def _visualize(self):
        # plt.figure()
        # plt.title("Grayscale Histogram")
        # plt.xlabel("Bins") # plt.ylabel("# of Pixels")
        # plt.plot(hist)
        # plt.xlim([0, 256])

        (fig, ax) = plt.subplots(1, 2, )
        ax[0].plot(self._hist_image)
        ax[0].set_title("self._hist_image")
        ax[0].set_xticks([])
        ax[0].set_yticks([])
        # display the magnitude image
        ax[1].plot(self._hist_template)
        ax[1].set_title("self._hist_template")
        ax[1].set_xticks([])
        ax[1].set_yticks([])
        # show our plots
        plt.show()

    def _compare(self):
        return self._comparison_method(
            self._hist_image.flatten(), self._hist_template.flatten())

    def __call__(
        self,
        image,
        template,
        vis=False,
    ):
        self._image = image

        self._template = template
        if len(self._image.shape) > 2:
            self._image = cv2.cvtColor(self._image, cv2.COLOR_BGR2GRAY)

        if len(self._template.shape) > 2:
            self._template = cv2.cvtColor(self._template, cv2.COLOR_BGR2GRAY)

        self._hist_image = cv2.calcHist(
            [self._image], [0], None, [self._hist_size], self._hist_range)
        self._hist_template = cv2.calcHist(
            [self._template], [0], None, [self._hist_size], self._hist_range)

        cv2.normalize(
            self._hist_image,
            self._hist_image,
            alpha=0,
            beta=255,
            norm_type=self._norm_type
        )

        cv2.normalize(
            self._hist_template,
            self._hist_template,
            0,
            255,
            self._norm_type
        )

        if vis:
            self._visualize()

        if self._debug:
            cv2.imshow("self._image", self._image)
            cv2.waitKey(0)
            cv2.imshow("self._template", self._template)
            cv2.waitKey(0)

        return self._compare()


class HueSaturationDifference:
    def __init__(
        self,
        N_h=5,
        N_w=10,
        h_magic_number=179,
        s_magic_number=255
    ):

        self._N_h = N_h
        self._N_w = N_w
        self._total_bins = self._N_h * self._N_w

        self._h_magic_number = h_magic_number
        self._s_magic_number = s_magic_number

    def _get_hsd_score(self):
        hsd_score = np.sum(
            self._image_h_s_means - self._template_h_s_means, axis=0
        )

        h_denom = self._h_magic_number * self._total_bins s_denom = self._s_magic_number * self._total_bins hsd_score = (hsd_score[0] * hsd_score[1])/(h_denom*s_denom)
        return abs(hsd_score)

    def __call__(self, image, template):
        self._template = cv2.cvtColor(template, cv2.COLOR_BGR2HSV)
        _, self._template_width, _ = self._template.shape

        self._image = imutils.resize(image, width=self._template_width)
        self._image = cv2.cvtColor(self._image, cv2.COLOR_BGR2HSV)

        self._image_h_s_means = get_hsv_means_for_windows(self._image,
                                                          N_h=self._N_h,
                                                          N_w=self._N_w,
                                                          debug=False)

        self._template_h_s_means = get_hsv_means_for_windows(self._template,
                                                             N_h=self._N_h,
                                                             N_w=self._N_w,
                                                             debug=False)
        return self._get_hsd_score()


class BrightspotDetector:
    def __init__(
        self,
        preprocess=[]
    ):

        self._preprocess = preprocess

    def __call__(
        self,
        image,
        percentage_limit=0.02,
        debug=False
    ):
        self._image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        self._blurred = cv2.GaussianBlur(self._image, (11, 11), 0)
        self._size = self._image.shape[0] * self._image.shape[1]
        self._percentage_limit = percentage_limit
        self._blob_size_limit = int(self._size * self._percentage_limit)

        # threshold the image to reveal light regions in the
        # blurred image

        self._thresh = cv2.threshold(
            self._blurred, 238, 255, cv2.THRESH_BINARY)[1]

        self._thresh = cv2.erode(self._thresh, None, iterations=2)
        self._thresh = cv2.dilate(self._thresh, None, iterations=4)

        # perform a connected component analysis on the thresholded
        # image, then initialize a mask to store only the "large"
        # components
        labels = measure.label(self._thresh, background=0)
        if debug:
            cv2.imshow("self._thresh", self._thresh)
            cv2.waitKey(0)
            print(labels)
        #mask = np.zeros(self._thresh.shape, dtype="uint8")
        # loop over the unique components
        for label in np.unique(labels):
            # if this is the background label, ignore it
            if label == 0:
                continue
            # otherwise, construct the label mask and count the
            # number of pixels
            labelMask = np.zeros(self._thresh.shape, dtype="uint8")
            labelMask[labels == label] = 255
            numPixels = cv2.countNonZero(labelMask)
            # if the number of pixels in the component is sufficiently
            # large, then add it to our mask of "large blobs"
            if numPixels > self._blob_size_limit:
                raise RuntimeError("Image has too many bright spots")

        return True


class StructuralSimilarity():
    def __init__(
        self,
        debug=False
    ):

        self._debug = debug

    def __call__(
        self,
        image,
        template,
        width=450,
        width_with_padding=600
    ):
        self._image = pad(
            imutils.resize(
                load_image(image),
                width=width
            ),
            width_with_padding
        )
        self._template = pad(
            imutils.resize(
                load_image(template),
                width=width
            ),
            width_with_padding
        )

        self._image_gray = cv2.cvtColor(
            self._image,
            cv2.COLOR_BGR2GRAY
        )

        self._template_gray = cv2.cvtColor(
            self._template,
            cv2.COLOR_BGR2GRAY
        )

        if self._debug:
            cv2.imshow("image", self._image_gray)
            cv2.waitKey(0)
            cv2.imshow("template", self._template_gray)
            cv2.waitKey(0)
        score, diff = metrics.structural_similarity(
            self._image_gray,
            self._template_gray,
            full=True
        )

        return score


class ColoCoherenceVectors():
    def __init__(
        self,
        debug=True
    ):

        self._debug = debug

    def quantize_image(
        image,
        clusters=6
    ):

        arr = image.reshape((-1, 3))
        kmeans = KMeans(n_clusters=clusters).fit(arr)
        labels = kmeans.labels_
        centers = kmeans.cluster_centers_
        less_colors = centers[labels].reshape(image.shape).astype('uint8')

        return less_colors

    def __call__(
        self,
        image=None,
        bins=6,
        area_low_ratio=.10,
        area_high_ratio=.30
    ):

        self._image = load_image(image)
        self._image = cv2.blur(self._image, (3, 3))

        self._bins = bins
        self._image_hsv = cv2.cvtColor(
            self._image,
            cv2.COLOR_BGR2HSV
        )

        #hue, saturation, value = cv2.split(self._image_hsv)
        # self._quantized_hue = quantize_image(
        #    hue,
        #    clusters=self._bins)

        self._quantized_image = quantize_image(
            self._image_hsv,
            clusters=self._bins
        )

        labeled_image, nl = measure.label(
            self._quantized_image,
            return_num=True
        )

        regions = measure.regionprops(labeled_image)
        masks = []
        bbox = []
        list_of_index = []

        area_image = labeled_image.shape[0] * labeled_image.shape[1]
        area_low = area_image*area_low_ratio
        area_high = area_image*area_high_ratio
        for num, x in enumerate(regions):
            area = x.area
            area_low = area*area_low_ratio
            area_high = area*area_high_ratio

            if (num!=0 and area_low < area):
                bbox.append(regions[num].bbox)
                list_of_index.append(num)

        #for box, mask in zip(bbox, masks):
        for box in bbox:
            print(f"{box=}")
            image =self._image_hsv[
                box[0]:box[2], 
                box[1]:box[3],
                :
            ] 
            #red = self._image_hsv[:, :, 0][
            #    box[0]:box[2], box[1]:box[3]]

            #green = self._image_hsv[:, :, 1][
            #    box[0]:box[2], box[1]:box[3]]

            #blue = self._image_hsv[:, :, 2][
            #    box[0]:box[2], box[1]:box[3]]

            #image = np.dstack([red, green, blue])
            imshow(image)
            plt.show()


        # region_df = pd.DataFrame(skimage.measure.regionprops_table(
        #    labeled_image,
        #    self._quantized_image,
        #    properties=['area', 'mean_intensity']
        # )
        # )
        # print(region_df)

        # filtered_areas_df = region_df[
        #    (region_df["area"] >= area*area_low_ratio) &
        #    (region_df["area"] <= area*area_high_ratio) &
        #    (region_df["mean_intensity"] >= 100)
        # ]

        #list_of_index = filtered_areas_df.index.values.tolist()

        # print(rgb_mask.shape)
        # for x in list_of_index[1:]:

        #    sl = ndimage.find_objects(labeled_image == x)
        #    region_selected = labeled_image[sl[0]]
        #    #cv2.imshow("region_selected",region_selected.astype("uint8"))
        #    #cv2.waitKey(0)
        #    imshow(region_selected)
        #    plt.show()

        #for x in list_of_index[1:]:

        #    print((labeled_image == x).shape)

        #red  =  painting[:,:,0] * rgb_mask
        #green = painting[:,:,1] * rgb_mask
        #blue  = painting[:,:,2] * rgb_mask
        #image = np.dstack([red, green, blue])
        # imshow(image)

        #print(region_selected)
        #print(labeled_image.shape)
        return nl





from sabhi_utils.image_utils.utils import load_image
from sabhi_utils.image_utils.utils import get_image_area
from sabhi_utils.image_utils.utils import quantize_image

from matplotlib import pyplot as plt
from skimage.io import imshow
import skimage.measure as measure
from scipy import ndimage
import cv2
import math
import pandas as pd
import time


def plot_figures(figures, nrows=None, ncols=None):
    """Plot a dictionary of figures.

    Parameters
    ----------
    figures : <title, figure> dictionary
    ncols : number of columns of subplots wanted in the display
    nrows : number of rows of subplots wanted in the figure
    """

    if (not nrows) and (not ncols):

        half_len = math.sqrt(len(figures))
        nrows = math.floor(half_len)
        ncols = math.ceil(half_len)

    fig, axeslist = plt.subplots(ncols=ncols, nrows=nrows)
    for ind, title in zip(range(len(figures)), figures):
        axeslist.ravel()[ind].imshow(figures[title])
        axeslist.ravel()[ind].set_title(title)
        axeslist.ravel()[ind].set_axis_off()
    plt.tight_layout()  # optional


def label_components(
    image,
):

    # im, number_of_objects = ndimage.label(image)
    # blobs = ndimage.find_objects(im)

    labeled_image = measure.label(
        image
    )

    return labeled_image


def get_region_thresholds(
    area,
    low_ratio=None,
    high_ratio=None
):
    region_low_thresh = area*low_ratio
    region_high_thresh = area*high_ratio

    return region_low_thresh, region_high_thresh


class CCVQuantizedImage:

    def __init__(
        self,
        debug=False
    ):
        self._debug = debug

    def _get_quantized_channels(
        self,
        channels
    ):
        quantized_channels = {}

        for channel_name, channel_array in channels.items():
            quantized_channels[channel_name] = self._quantize_channel(
                channel_array
            )
        return quantized_channels

    def _get_connected_components(
        self,
        channels
    ):
        connected_components = {}

        for channel_name, channel_array in channels.items():

            connected_components[
                channel_name] = label_components(
                channels[channel_name]
            )
        return connected_components

    def _get_connected_blobs_skimage(
        self,
        connected_components,
        list_of_index,
        key
    ):
        blobs = connected_components[key]

        connected_regions = {}
        for x in list_of_index[key]:
            sl = ndimage.find_objects(blobs == x)
            region_selected = self._image[sl[0]]
            connected_regions[x] = region_selected
        return connected_regions

    def _is_reasonable_area(
        self,
        region
    ):
        return self._region_low_thresh < region.area < self._region_high_thresh

    def _is_reasonable_intensity(
        self,
        region
    ):
        return region.mean_intensity > self._intensity_threshold

    def _area_and_intensity(
        self,
        region
    ):
        return all(
            [
                self._is_reasonable_area(region)  # ,
                # self._is_reasonable_intensity(region),
            ]
        )

    def _filtered_region_list_index(
        self,
        labeled_image,
        criteria
    ):

        regions = measure.regionprops(
            labeled_image
        )
        list_of_index = []
        for region in regions:
            if criteria(region):
                list_of_index.append(region.label)

        return list_of_index

    def _filtered_index_list(
        self,
        connected_components,
        criteria
    ):
        filtered_cc = {}
        for key, labeled_image in connected_components.items():
            filtered_cc[key] = self._filtered_region_list_index(
                labeled_image,
                criteria
            )
        return filtered_cc

    def __call__(
        self,
        image=None,
        bins=6,
        blur=True,
        area_low_ratio=.02,
        area_high_ratio=.4,
        intensity_threshold=150
    ):

        self._bins = bins
        self._image = load_image(image)
        self._image_area = get_image_area(self._image)
        if blur:
            self._image = cv2.blur(self._image, (5, 5))

        (self._region_low_thresh,
         self._region_high_thresh) = get_region_thresholds(
            self._image_area,
            area_low_ratio,
            area_high_ratio
        )

        self._intensity_threshold = intensity_threshold

        self._image_rgb = cv2.cvtColor(
            self._image,
            cv2.COLOR_BGR2RGB
        )

        self._image_hsv = cv2.cvtColor(
            self._image_rgb,
            cv2.COLOR_BGR2HSV
        )

        quantized_image = quantize_image(
            self._image_hsv,
            clusters=self._bins
        )

        (h_quantized_image,
         s_quantized_image,
         v_quantized_image) = cv2.split(quantized_image)

        quantized_image_channels = {
            "h_quantized_image": h_quantized_image,
            "s_quantized_image": s_quantized_image,
            "v_quantized_image": v_quantized_image
        }

        connected_components_quantized_image = self._get_connected_components(
            quantized_image_channels
        )

        start_time = time.perf_counter()

        image_indices = self._filtered_index_list(
            connected_components_quantized_image,
            criteria=self._area_and_intensity
        )

        elapsed_time = time.perf_counter() - start_time
        print(f"LOOP {elapsed_time=}")

        if self._debug:
            images = {
              "hsv": self._image_hsv,
              "hsv_quantized": quantized_image["hsv"],
              "h_quantized_image": h_quantized_image,
              "s_quantized_image": s_quantized_image,
              "v_quantized_image": v_quantized_image,
            }

            plot_figures(images,2,2)
            plt.show()

            print(image_indices)
            connected_regions = self._get_connected_blobs_skimage(
                connected_components_quantized_image,
                image_indices,
                key="s_quantized_image"
            )

            plot_figures(connected_regions)
            plt.show()

            connected_regions = self._get_connected_blobs_skimage(
                connected_components_quantized_image,
                image_indices,
                key="h_quantized_image"
            )
            plot_figures(connected_regions)
            plt.show()
        return None


class CCVQuantizedChannels:

    def __init__(
        self,
        debug=False
    ):
        self._debug = debug

    def _quantize_image(
        self,
        channel,
    ):
        return quantize_image(
            channel,
            clusters=self._bins)

    def _label_components(
        self,
        image,
    ):

        # im, number_of_objects = ndimage.label(image)
        # blobs = ndimage.find_objects(im)

        labeled_image = measure.label(
            image
        )

        return labeled_image

    def _get_quantized_channels(
        self,
        channels
    ):
        quantized_channels = {}

        for channel_name, channel_array in channels.items():
            quantized_channels[channel_name] = self._quantize_channel(
                channel_array
            )
        return quantized_channels

    def _get_connected_components_(
        self,
        channels
    ):
        connected_components = {}

        for channel_name, channel_array in channels.items():

            connected_components[
                channel_name] = self._label_components(
                channels[channel_name]
            )
        return connected_components

    def _get_connected_blobs_skimage(
        self,
        connected_components,
        list_of_index,
        key
    ):
        blobs = connected_components[key]
        # for i, j in enumerate(blobs):
        #    blob[i] = self._image[j]

        connected_regions = {}
        for x in list_of_index[key]:
            sl = ndimage.find_objects(blobs == x)
            region_selected = self._image[sl[0]]
            connected_regions[x] = region_selected
        return connected_regions

    def _is_reasonable_area(
        self,
        region
    ):
        return self._region_low_thresh < region.area < self._region_high_thresh

    def _is_reasonable_intensity(
        self,
        region
    ):
        return region.mean_intensity > self._intensity_threshold

    def _area_and_intensity(
        self,
        region
    ):
        return all(
            [
                self._is_reasonable_area(region)  # ,
                # self._is_reasonable_intensity(region),
            ]
        )

    # def _filter_regions(
    #    self,
    #    labeled_image,
    #    quantized_image
    #    criteria
    # ):
    #    region_df = pd.DataFrame(
    #        skimage.measure.regionprops_table(
    #       labeled_image,
    #       quantized_image,
    #       properties=['area', 'mean_intensity']
    #    )
    #    )

    def _filter_list_index(
        self,
        labeled_image,
        image,
    ):
        region_df = pd.DataFrame(
            measure.regionprops_table(
                labeled_image,
                image,
                properties=['label', 'area', 'mean_intensity']
            )
        )
        filtered_areas_df = region_df[
            (region_df["area"] >= self._region_low_thresh) &
            (region_df["area"] <= self._region_high_thresh)
        ]

        return filtered_areas_df["label"].tolist()

    def _filter_connected_components(
        self,
        connected_components,
        criteria
    ):
        filtered_cc = {}
        for key, labeled_image in connected_components.items():
            filtered_cc[key] = self._filter_list_index(
                labeled_image,
                self._image
            )
        return filtered_cc

    def __call__(
        self,
        image=None,
        bins=6,
        blur=True,
        area_low_ratio=.02,
        area_high_ratio=.4,
        intensity_threshold=150
    ):

        self._bins = bins
        self._image = load_image(image)
        self._image_area = self._image.shape[0] * self._image.shape[1]
        if blur:
            self._image = cv2.blur(self._image, (7, 7))

        self._region_low_thresh = self._image_area*area_low_ratio
        self._region_high_thresh = self._image_area*area_high_ratio
        self._intensity_threshold = intensity_threshold

        self._image_rgb = cv2.cvtColor(
            self._image,
            cv2.COLOR_BGR2RGB
        )

        self._image_hsv = cv2.cvtColor(
            self._image_rgb,
            cv2.COLOR_BGR2HSV
        )

        quantized_image = self._quantize_image(self._image_hsv)

        (h_quantized_image,
         s_quantized_image,
         v_quantized_image) = cv2.split(quantized_image)

        quantized_image_channels = {
            "h_quantized_image": h_quantized_image,
            "s_quantized_image": s_quantized_image,
            "v_quantized_image": v_quantized_image
        }

        connected_components_quantized_image = self._get_connected_components_(
            quantized_image_channels
        )

        quantized_image_indices = self._filter_connected_components(
            connected_components_quantized_image,
            criteria=self._area_and_intensity
        )

        h, s, v = cv2.split(self._image_hsv)

        quantized_h = quantize_image(h)
        quantized_s = quantize_image(s)
        quantized_v = quantize_image(v)

        quantized_channels = {
            "h_quantized": quantized_h,
            "s_quantized": quantized_s,
            "v_quantized": quantized_v
        }

        connected_components_quantized_channels = self._get_connected_components_(
            quantized_channels
        )

        quantized_channels_indices = self._filter_connected_components(
            connected_components_quantized_channels,
            criteria=self._area_and_intensity
        )

        if self._debug:
            images = {
                "hsv": self._image_hsv,
                "hsv_quantized": quantized_image["hsv"],
                "h_quantized_image": h_quantized_image,
                "s_quantized_image": s_quantized_image,
                "v_quantized_image": v_quantized_image,
                "h_quantized": quantized_h,
                "s_quantized": quantized_s,
                "v_quantized": quantized_v
            }

            plot_figures(images)
            plt.show()
            # key = "s_quantized"

            connected_regions = self._get_connected_blobs_skimage(
                connected_components_quantized_image,
                quantized_image_indices,
                key="h_quantized_image"
            )
            print(connected_regions)

            plot_figures(connected_regions, 5, 5)
            plt.show()

            connected_regions = self._get_connected_blobs_skimage(
                connected_components_quantized_channels,
                quantized_channels_indices,
                key="h_quantized"
            )
            print(connected_regions)

            plot_figures(connected_regions, 2, 2)
            plt.show()

        return None



import numpy as np
from sabhi_utils.image_utils.processors import Closer
from sabhi_utils.image_utils.processors import EdgeDetector
from sabhi_utils.image_utils.processors import Opener
from collections import defaultdict
import itertools
import cv2
import time

from sklearn.cluster import KMeans


class HoughLineCornerDetector:
    def __init__(
        self,
        rho_acc=2,
        theta_acc=360,
        thresh=100,
        output_process=True
    ):

        self.rho_acc = rho_acc
        self.theta_acc = theta_acc
        self.thresh = thresh
        self.output_process = output_process
        self._preprocessor = [
            Closer(output_process=output_process),
            EdgeDetector(output_process=output_process),
            #Opener(output_process=output_process)
        ]

    def _get_hough_lines(self):
        """
        Extract straight lines from image using Hough Transform.

        Returns
        ----------
        array containing rho and theta of lines (Hess Norm formulation)
        """
        image = cv2
        lines = cv2.HoughLines(
            self._image, self.rho_acc, np.pi / self.theta_acc, self.thresh
        )

        return lines

    def _draw_hough_lines(
        self,
        image,
        lines,
        file_prefix=str(time.time())
    ):
        hough_line_output = image
        n = self._n

        for line in lines:
            rho, theta = line[0]
            a, b = np.cos(theta), np.sin(theta)
            x0, y0 = a * rho, b * rho
            x1 = int(x0 + n * (-b))
            y1 = int(y0 + n * (a))
            x2 = int(x0 - n * (-b))
            y2 = int(y0 - n * (a))

            cv2.line(hough_line_output, (x1, y1), (x2, y2), (0, 0, 255), 2)

        cv2.imwrite("output/"+file_prefix+"hough_line.jpg", hough_line_output)

    def _segment_by_angle_kmeans(self, k=2, **kwargs):
        """Groups lines based on angle with k-means.

        Uses k-means on the coordinates of the angle on the unit circle
        to segment `k` angles inside `lines`.
        """

        lines = self._lines

        # Define criteria = (type, max_iter, epsilon)
        default_criteria_type = (
            cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER
        )

        criteria = kwargs.get("criteria", (default_criteria_type, 10, 1.0))
        flags = kwargs.get("flags", cv2.KMEANS_RANDOM_CENTERS)
        attempts = kwargs.get("attempts", 10)

        # returns angles in [0, pi] in radians
        angles = np.array([line[0][1] for line in lines])
        # multiply the angles by two and find coordinates of that angle
        pts = np.array(
            [[np.cos(2 * angle), np.sin(2 * angle)] for angle in angles],
            dtype=np.float32,
        )

        # run kmeans on the coords
        labels, centers = cv2.kmeans(
            pts, k, None, criteria, attempts, flags)[1:]
        labels = labels.reshape(-1)  # transpose to row vec

        # segment lines based on their kmeans label
        segmented = defaultdict(list)
        for i, line in zip(range(len(lines)), lines):
            segmented[labels[i]].append(line)
        segmented = list(segmented.values())
        return segmented

    def _get_intersections(self, k=2):
        """Finds the intersections between groups of lines."""
        segmented_lines = self._segmented_lines
        intersections = []

        group_lines = list(itertools.product(
            segmented_lines[0], segmented_lines[1]))

        for line_i, line_j in group_lines:
            int_point = self._intersection(line_i, line_j)
            intersections.append(int_point)


        return intersections

    def _find_quadrilaterals(self):
        X = np.array([[point[0][0], point[0][1]]
                      for point in self._intersections])

        kmeans = KMeans(
            n_clusters=4,
            init="k-means++",
            max_iter=100,
            n_init=10,
            random_state=0
        ).fit(X)

        if self.output_process:
            self._draw_quadrilaterals(self._lines, kmeans)

        return [[center.tolist()] for center in kmeans.cluster_centers_]

    def _draw_quadrilaterals(
        self,
        lines,
        kmeans,
        file_prefix=str(time.time())
    ):
        grouped_output = self._get_color_image()
        n = self._n

        for idx, line in enumerate(lines):
            rho, theta = line[0]
            a, b = np.cos(theta), np.sin(theta)
            x0, y0 = a * rho, b * rho
            x1 = int(x0 + n * (-b))
            y1 = int(y0 + n * (a))
            x2 = int(x0 - n * (-b))
            y2 = int(y0 - n * (a))

            cv2.line(grouped_output, (x1, y1), (x2, y2), (0, 0, 255), 2)

        for point in kmeans.cluster_centers_:
            x, y = point

            cv2.circle(grouped_output, (int(x), int(y)), 5, (255, 255, 255), 5)

        cv2.imwrite(
            "output/"+str(file_prefix)+"grouped_quads.jpg",
            grouped_output
        )

    def _get_angle_between_lines(self, line_1, line_2):
        rho1, theta1 = line_1[0]
        rho2, theta2 = line_2[0]
        # x * cos(theta) + y * sin(theta) = rho
        # y * sin(theta) = x * (- cos(theta)) + rho
        # y = x * (-cos(theta) / sin(theta)) + rho
        m1 = -(np.cos(theta1) / np.sin(theta1))
        m2 = -(np.cos(theta2) / np.sin(theta2))
        return abs(np.arctan(abs(m2 - m1) / (1 + m2 * m1))) * (180 / np.pi)

    def _intersection(self, line1, line2):
        """Finds the intersection of two lines given in Hesse normal form.

        Returns closest integer pixel locations.
        See https://stackoverflow.com/a/383527/5087436
        """
        rho1, theta1 = line1[0]
        rho2, theta2 = line2[0]

        A = np.array(
            [[np.cos(theta1), np.sin(theta1)], [
                np.cos(theta2), np.sin(theta2)]]
        )

        b = np.array([[rho1], [rho2]])
        x0, y0 = np.linalg.solve(A, b)
        x0, y0 = int(np.round(x0)), int(np.round(y0))
        return [[x0, y0]]

    def _draw_intersections(
        self,
        image,
        intersections,
        file_prefix=str(time.time())
    ):
        intersection_point_output = image
        n = self._n

        for line in self._lines:
            rho, theta = line[0]
            a = np.cos(theta)
            b = np.sin(theta)
            x0 = a * rho
            y0 = b * rho
            x1 = int(x0 + n * (-b))
            y1 = int(y0 + n * (a))
            x2 = int(x0 - n * (-b))
            y2 = int(y0 - n * (a))

            cv2.line(intersection_point_output,
                     (x1, y1), (x2, y2), (0, 0, 255), 2)

        for point in intersections:
            x, y = point[0]

            cv2.circle(intersection_point_output,
                       (x, y), 5, (255, 255, 127), 5)

        cv2.imwrite("output/"+file_prefix+"intersection_point_output.jpg",
                    intersection_point_output)

    def _get_color_image(self):
        return cv2.cvtColor(self._image.copy(), cv2.COLOR_GRAY2RGB)

    def __call__(
        self,
        image,
        debug=False
    ):
        # Step 1: Process for edge detection
        self._image = image
        self._n = max(self._image.shape[0], self._image.shape[1])

        for processor in self._preprocessor:
            self._image = processor(self._image)

        # Step 2: Get hough lines
        self._lines = self._get_hough_lines()
        self._segmented_lines = self._segment_by_angle_kmeans()

        # Step 3: Get intersection points
        self._intersections = self._get_intersections()

        # Step 4: Get Quadrilaterals

        self._quadrilaterals = self._find_quadrilaterals()
        if self.output_process:
            self._draw_intersections(
                image,
                self._intersections
            )

            self._draw_hough_lines(
                image,
                self._lines)

        return self._quadrilaterals




from .utils import get_aspect_ratio_check

from sklearn.cluster import KMeans

import numpy as np

import cv2

from collections import defaultdict

import itertools

from .utils import get_bounding_box_points


def get_hough_lines(
    image,
    rho_acc=1,
    theta_acc=180,
    thresh=70,
):
    """
    Extract straight lines from image using Hough Transform.

    Returns
    ----------
    array containing rho and theta of lines (Hess Norm formulation)
    """
    lines = cv2.HoughLines(
        image,
        rho_acc,
        np.pi / theta_acc,
        thresh
    )

    return lines


def segment_by_angle_kmeans(
    lines=None,
    k=2,
    **kwargs
):
    """Groups lines based on angle with k-means.

    Uses k-means on the coordinates of the angle on the unit circle
    to segment `k` angles inside `lines`.
    """
# Define criteria = (type, max_iter, epsilon)
    default_criteria_type = (
        cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER
    )

    criteria = kwargs.get("criteria", (default_criteria_type, 10, 1.0))
    flags = kwargs.get("flags", cv2.KMEANS_RANDOM_CENTERS)
    attempts = kwargs.get("attempts", 5)

    # returns angles in [0, pi] in radians
    angles = np.array(
        [line[0][1] for line in lines]
    )
# multiply the angles by two and find coordinates of that angle
    pts = np.array(
        [[np.cos(2 * angle), np.sin(2 * angle)] for angle in angles],
        dtype=np.float32,
    )

    # run kmeans on the coords
    labels, centers = cv2.kmeans(
        pts, k, None, criteria, attempts, flags)[1:]
    labels = labels.reshape(-1)  # transpose to row vec

    # segment lines based on their kmeans label
    # segmented = list()
    # for i, line in zip(range(len(lines)), lines):
    #    segmented.append(line)
    segmented = defaultdict(list)
    for i, line in zip(range(len(lines)), lines):
        segmented[labels[i]].append(line)

    segmented = list(segmented.values())
    return segmented


def intersection(
    line1,
    line2
):
    """Finds the intersection of two lines given in Hesse normal form.

    Returns closest integer pixel locations.
    See https://stackoverflow.com/a/383527/5087436
    """
    rho1, theta1 = line1[0]
    rho2, theta2 = line2[0]

    A = np.array(
        [[np.cos(theta1), np.sin(theta1)], [
            np.cos(theta2), np.sin(theta2)]]
    )

    b = np.array([[rho1], [rho2]])
    x0, y0 = np.linalg.solve(A, b)
    x0, y0 = int(np.round(x0)), int(np.round(y0))
    return [x0, y0]


def get_intersections(
    lines=None,
):
    """Finds the intersections between groups of lines."""
    intersections = []

    group_lines = list(
        itertools.product(lines[0], lines[1])
    )

    for line_i, line_j in group_lines:
        int_point = intersection(line_i, line_j)
        intersections.append(int_point)

    return np.array(intersections)


def get_optimal_cluster(
    points=None,
    max_k=20
):
    inertia_list = []
    for n_clusters in range(1, max_k+1):
        if len(points) >= n_clusters:
            model = KMeans(
                n_clusters=n_clusters,
                init="k-means++",
                max_iter=100,
                n_init=10,
                random_state=0
            ).fit(points)

            inertia_list.append(model.inertia_)

    k = [i*100 for i in np.diff(inertia_list, 2)].index(
        min(
            [i*100 for i in np.diff(inertia_list, 2)]
        )
    )

    return k


def find_cluster_centers(
    intersections,
    n_clusters=None
):

    if n_clusters is None:
        n_clusters = get_optimal_cluster(points=intersections)

    model = KMeans(
        n_clusters=n_clusters,
        init="k-means++",
        max_iter=100,
        n_init=10,
        random_state=0
    ).fit(intersections)

    return model.cluster_centers_


def unit_vector(vector):
    """ Returns the unit vector of the vector.  """
    norm = np.linalg.norm(vector)
    response = vector / norm
    return response


def angle_between(v1, v2):
    """ Returns the angle in radians between vectors 'v1' and 'v2'::

            >>> angle_between((1, 0, 0), (0, 1, 0))
            1.5707963267948966
            >>> angle_between((1, 0, 0), (1, 0, 0))
            0.0
            >>> angle_between((1, 0, 0), (-1, 0, 0))
            3.141592653589793
    """
    v1_u = unit_vector(v1)
    v2_u = unit_vector(v2)
    if np.isnan(v1_u).any() or np.isnan(v2_u).any():
        print(v1, v2)
        return None

    return np.arccos(np.clip(np.dot(v1_u, v2_u), -1.0, 1.0))


def check_quadrilateral_angles(
    rect,
    threshold=1.39
):
    p1, p2, p3, p4 = rect

    l1 = p2-p1

    l2 = p3-p2

    l3 = p4-p3

    l4 = p1-p4

    pair_lines = [
        (l1, l2),
        (l2, l3),
        (l3, l4)
    ]

    sum_of_angles = 0
    for line_1, line_2 in pair_lines:
        angle = (
            angle_between(
                line_1,
                line_2
            )
        )

        if angle is None or angle < threshold:
            return None
        sum_of_angles = sum_of_angles + angle

        if (2*np.pi - sum_of_angles) < threshold:
            return None

    return True


def get_max_width(
    rect
):
    (tl, tr, br, bl) = rect
    # compute the width of the new image, which will be the
    # maximum distance between bottom-right and bottom-left
    # x-coordiates or the top-right and top-left x-coordinates
    widthA = np.sqrt(((br[0] - bl[0]) ** 2) + ((br[1] - bl[1]) ** 2))
    widthB = np.sqrt(((tr[0] - tl[0]) ** 2) + ((tr[1] - tl[1]) ** 2))
    maxWidth = max(int(widthA), int(widthB))

    return maxWidth


def get_max_height(
    rect
):
    (tl, tr, br, bl) = rect
    # compute the height of the new image, which will be the
    # maximum distance between the top-right and bottom-right
    # y-coordinates or the top-left and bottom-left y-coordinates
    heightA = np.sqrt(((tr[0] - br[0]) ** 2) + ((tr[1] - br[1]) ** 2))
    heightB = np.sqrt(((tl[0] - bl[0]) ** 2) + ((tl[1] - bl[1]) ** 2))
    maxHeight = max(int(heightA), int(heightB))

    return maxHeight


def check_aspect_ratio(
    rect,
    aspect_ratio=16/10
):

    response = None

    maxWidth = get_max_width(rect)

    maxHeight = get_max_height(rect)

    aspect_ratio_check = get_aspect_ratio_check(
        height=maxHeight,
        width=maxWidth,
        aspect_ratio=aspect_ratio,
        threshold=.05
    )

    if aspect_ratio_check:
        response = True

    return response


def check_quadrilateral(
    two_line_segments,
    checks=[
        check_quadrilateral_angles,
        check_aspect_ratio
    ]
):
    response = None
    pts = np.array(
        [point for segment in two_line_segments for point in segment]
    )

    rect = get_bounding_box_points(pts)

    unique = np.unique(rect, axis=0)

    if len(unique) < rect.shape[0]:
        return None
    for checker in checks:
        check_response = checker(rect)
        if check_response is None:
            return response

    return True


def get_convex_quadrilaterals(
    points,
    check=check_quadrilateral
):
    """
    Generate the convex quadrilaterals among points on a 2D Surface (np.array).
    """
    points = [point for point in points]
    # points = map(np.array, points)
    segments = itertools.combinations(points, 2)

    for s0, s1 in itertools.combinations(segments, 2):
        response = check((s0, s1))
        if response is not None:
            yield s0, s1


def get_points_from_quad(quad):
    pts = np.array(
        [point for segment in quad for point in segment]
    )
    rect = get_bounding_box_points(pts)
    return rect


def get_convex_quad_array(convex_quads):
    return map(get_points_from_quad, convex_quads)


def get_bounding_box_quad(
    convex_quadrilateals=None
):
    convex_quads_pts = get_convex_quad_array(
        convex_quadrilateals
    )
    max_quad = []
    max_area = 0
    for quad in convex_quads_pts:
        area = get_max_width(quad)*get_max_height(quad)
        if area > max_area:
            max_area = area
            max_quad = quad

    return max_quad



import cv2
import numpy as np
import imutils


class TemplateMatching:

    def __init__(
        self,
        method=cv2.TM_CCOEFF_NORMED,
        matching_threshold=0.75,
        scales=np.linspace(0.2, 1.0, 20)[::-1],
        debug=False
    ):

        self._method = method
        self._matching_threshold = matching_threshold
        self._scales = scales
        self._debug = debug

    def _multi_scale_match_template(self, found=None):
        image = self._image.copy()
        for scale in self._scales:
            resized_image = imutils.resize(
                image,
                width=int(image.shape[1] * scale)
            )

            r = image.shape[1] / float(resized_image.shape[1])

            if resized_image.shape[0] < self._template_height or resized_image.shape[1] < self._template_width:
                break

            result = cv2.matchTemplate(
                resized_image,
                self._template,
                method=self._method
            )

            match_locations = np.where(result >= self._matching_threshold)
            (_, maxVal, _, maxLoc) = cv2.minMaxLoc(result)

            if self._debug:
                # draw a bounding box around the detected region
                clone = resized_image
                cv2.rectangle(clone, (maxLoc[0], maxLoc[1]),
                              (maxLoc[0] + self._template_width, maxLoc[1] + self._template_height), (0, 0, 255), 2)
                cv2.imshow("Visualize", clone)
                cv2.waitKey(0)

                #cv2.imshow("result", result)
                # cv2.waitKey(0)
            if found is None or maxVal > found[0]:
                found = (maxVal, maxLoc, r)
        return found

    def __call__(
        self,
        image,
        template,
        label="1"
    ):
        self._image = image
        self._template = template

        self._template_height, self._template_width = self._template.shape[:2]

        result = self._multi_scale_match_template()

        return result



#import face_recognition
import cv2
import os
import wget
import pandas as pd
from PIL import Image

from sabhi_utils.image_utils.utils import load_image
from pathlib import Path


def initialize_face_detector(detector_backend):
    home = str(Path.home())
    output_dir = os.path.join(
        home,
        '.sabhi_utils/weights/'
    )

    os.makedirs(output_dir, exist_ok=True)
    if detector_backend == 'ssd':

        # check required ssd model exists in the home/.sabhi_utils/weights folder

        # model structure
        if os.path.isfile(
            os.path.join(output_dir, 'deploy.prototxt')
        ) != True:

            print("deploy.prototxt will be downloaded...")

            url = "https://github.com/opencv/opencv/raw/3.4.0/samples/dnn/face_detector/deploy.prototxt"

            output = os.path.join(output_dir, 'deploy.prototxt')

            wget.download(url, output)

        # pre-trained weights
        if os.path.isfile(
            os.path.join(
                output_dir,
                'res10_300x300_ssd_iter_140000.caffemodel')
        ) != True:

            print("res10_300x300_ssd_iter_140000.caffemodel will be downloaded...")

            url = "https://github.com/opencv/opencv_3rdparty/raw/dnn_samples_face_detector_20170830/res10_300x300_ssd_iter_140000.caffemodel"

            output = os.path.join(
                output_dir,
                'res10_300x300_ssd_iter_140000.caffemodel'
            )

            wget.download(url, output)

        face_detector = cv2.dnn.readNetFromCaffe(
            home+"/.sabhi_utils/weights/deploy.prototxt",
            home+"/.sabhi_utils/weights/res10_300x300_ssd_iter_140000.caffemodel"
        )

    return face_detector


def detect_face(
    img,
    detector_backend='ssd',
    grayscale=False,
    tolerance=.88,
    enforce_detection=True,
    debug=False
):

    img = load_image(img)
    face_detector = initialize_face_detector(
        detector_backend=detector_backend
    )

    if detector_backend == 'ssd':
        ssd_labels = ["img_id", "is_face", "confidence",
                      "left", "top", "right", "bottom"]

        target_size = (300, 300)

        base_img = img.copy()  # we will restore base_img to img later
        original_size = img.shape

        img = cv2.resize(base_img, target_size)
        if debug:
            cv2.imshow("img", img)
            cv2.waitKey(0)
        #aspect_ratio_x = (original_size[1] / target_size[1])
        #aspect_ratio_y = (original_size[0] / target_size[0])
        imageBlob = cv2.dnn.blobFromImage(image=img)

        face_detector.setInput(imageBlob)
        detections = face_detector.forward()

        detections_df = pd.DataFrame(detections[0][0], columns=ssd_labels)

        detections_df = detections_df.loc[
            detections_df['confidence'] >= tolerance
        ]

        return detections_df


def num_of_faces(image):

    return len(detect_face(image).index)


class FaceRecognition:

    def __init__(
        self,
        debug=False
    ):

        self._debug = debug

    def __call__(
        self,
        image=None,
    ):
        self._image = image
        self._columns = [
            "confidence",
            "left",
            "top"
        ]

        self._detected_faces_df = detect_face(image)

        self._num_faces = len(self._detected_faces_df)

        if self._num_faces > 0:

            return {
                "num_of_faces": self._num_faces,
                self._columns[0]: self._detected_faces_df.loc[0, self._columns[0]],
                self._columns[1]: self._detected_faces_df.loc[0, self._columns[1]],
                self._columns[2]: self._detected_faces_df.loc[0, self._columns[2]]
            }



from sabhi_utils.image_utils.utils import load_image
from sabhi_utils.image_utils.utils import quantize_image

from matplotlib import pyplot as plt
from skimage.io import imshow
import skimage.measure as measure
from scipy import ndimage
import cv2
import math
import pandas as pd


def plot_figures(figures, nrows=None, ncols=None):
    """Plot a dictionary of figures.

    Parameters
    ----------
    figures : <title, figure> dictionary
    ncols : number of columns of subplots wanted in the display
    nrows : number of rows of subplots wanted in the figure
    """

    if (not nrows) and (not ncols):

        half_len = math.sqrt(len(figures))
        nrows = math.floor(half_len)
        ncols = math.ceil(half_len)

    fig, axeslist = plt.subplots(ncols=ncols, nrows=nrows)
    for ind, title in zip(range(len(figures)), figures):
        axeslist.ravel()[ind].imshow(figures[title])
        axeslist.ravel()[ind].set_title(title)
        axeslist.ravel()[ind].set_axis_off()
    plt.tight_layout()  # optional


class CCVQuantizedImage:

    def __init__(
        self,
        debug=False
    ):
        self._debug = debug

    def _quantize_image(
        self,
        channel,
    ):
        return quantize_image(
            channel,
            clusters=self._bins)

    def _label_components(
        self,
        image,
    ):

        #im, number_of_objects = ndimage.label(image)
        #blobs = ndimage.find_objects(im)

        labeled_image = measure.label(
            image
        )

        return labeled_image

    def _get_quantized_channels(
        self,
        channels
    ):
        quantized_channels = {}

        for channel_name, channel_array in channels.items():
            quantized_channels[channel_name] = self._quantize_channel(
                channel_array
            )
        return quantized_channels

    def _get_connected_components_(
        self,
        channels
    ):
        connected_components = {}

        for channel_name, channel_array in channels.items():

            connected_components[
                channel_name] = self._label_components(
                channels[channel_name]
            )
        return connected_components

    def _get_connected_blobs_skimage(
        self,
        connected_components,
        list_of_index,
        key
    ):
        blobs = connected_components[key]
        # for i, j in enumerate(blobs):
        #    blob[i] = self._image[j]

        connected_regions = {}
        for x in list_of_index[key]:
            sl = ndimage.find_objects(blobs == x)
            region_selected = self._image[sl[0]]
            connected_regions[x] = region_selected
        return connected_regions

    def _is_reasonable_area(
        self,
        region
    ):
        return self._region_low_thresh < region.area < self._region_high_thresh

    def _is_reasonable_intensity(
        self,
        region
    ):
        return region.mean_intensity > self._intensity_threshold

    def _area_and_intensity(
        self,
        region
    ):
        return all(
            [
                self._is_reasonable_area(region)  # ,
                # self._is_reasonable_intensity(region),
            ]
        )

    # def _filter_regions(
    #    self,
    #    labeled_image,
    #    quantized_image
    #    criteria
    # ):
    #    region_df = pd.DataFrame(
    #        skimage.measure.regionprops_table(
    #       labeled_image,
    #       quantized_image,
    #       properties=['area', 'mean_intensity']
    #    )
    #    )

    def _filter_list_index(
        self,
        labeled_image,
        image,
    ):
        region_df = pd.DataFrame(
            measure.regionprops_table(
                labeled_image,
                image,
                properties=['label', 'area', 'mean_intensity']
            )
        )
        filtered_areas_df = region_df[
            (region_df["area"] >= self._region_low_thresh) &
            (region_df["area"] <= self._region_high_thresh)
        ]

        return filtered_areas_df["label"].tolist()

    def _filter_connected_components(
        self,
        connected_components,
        criteria
    ):
        filtered_cc = {}
        for key, labeled_image in connected_components.items():
            filtered_cc[key] = self._filter_list_index(
                labeled_image,
                self._image
            )
        return filtered_cc

    def __call__(
        self,
        image=None,
        bins=20,
        blur=True,
        area_low_ratio=.02,
        area_high_ratio=1,
        intensity_threshold=150
    ):

        self._bins = bins
        self._image = load_image(image)
        self._image_area = self._image.shape[0] * self._image.shape[1]
        if blur:
            self._image = cv2.blur(self._image, (5, 5))

        self._region_low_thresh = self._image_area*area_low_ratio
        self._region_high_thresh = self._image_area*area_high_ratio
        self._intensity_threshold = intensity_threshold

        self._image_rgb = cv2.cvtColor(
            self._image,
            cv2.COLOR_BGR2RGB
        )

        self._image_hsv = cv2.cvtColor(
            self._image_rgb,
            cv2.COLOR_BGR2HSV
        )

        quantized_image = self._quantize_image(self._image_hsv)

        (h_quantized_image,
         s_quantized_image,
         v_quantized_image) = cv2.split(quantized_image)

        quantized_image_channels = {
            "h_quantized_image": h_quantized_image,
            "s_quantized_image": s_quantized_image,
            "v_quantized_image": v_quantized_image
        }

        connected_components_quantized_image = self._get_connected_components_(
            quantized_image_channels
        )

        quantized_image_indices = self._filter_connected_components(
            connected_components_quantized_image,
            criteria=self._area_and_intensity
        )

        h, s, v = cv2.split(self._image_hsv)

        quantized_h = quantize_image(h)
        quantized_s = quantize_image(s)
        quantized_v = quantize_image(v)

        quantized_channels = {
            "h_quantized": quantized_h,
            "s_quantized": quantized_s,
            "v_quantized": quantized_v
        }

        connected_components_quantized_channels = self._get_connected_components_(
            quantized_channels
        )

        quantized_channels_indices = self._filter_connected_components(
            connected_components_quantized_channels,
            criteria=self._area_and_intensity
        )

        if self._debug:
            images = {
                # "hsv": self._image_hsv,
                # "hsv_quantized": quantized_image["hsv"],
                "h_quantized_image": h_quantized_image,
                "s_quantized_image": s_quantized_image,
                "v_quantized_image": v_quantized_image,
                "h_quantized": quantized_h,
                "s_quantized": quantized_s,
                "v_quantized": quantized_v
            }

            plot_figures(images)
            plt.show()
            #key = "s_quantized"

            connected_regions = self._get_connected_blobs_skimage(
                connected_components_quantized_image,
                quantized_image_indices,
                key="h_quantized_image"
            )
            print(connected_regions)

            plot_figures(connected_regions, 5, 5)
            plt.show()

            connected_regions = self._get_connected_blobs_skimage(
                connected_components_quantized_channels,
                quantized_channels_indices,
                key="h_quantized"
            )
            print(connected_regions)

            plot_figures(connected_regions, 2, 2)
            plt.show()
        # get_conneted_components

        # get_filtered_component_masks
        # compare_extracted_masks_with_ground_truth
        # get_similarity_score for each channel

        return None


class CCVQuantizedChannels:

    def __init__(
        self,
        debug=False
    ):
        self._debug = debug

    def _quantize_image(
        self,
        channel,
    ):
        return quantize_image(
            channel,
            clusters=self._bins)

    def _label_components(
        self,
        image,
    ):

        #im, number_of_objects = ndimage.label(image)
        #blobs = ndimage.find_objects(im)

        labeled_image = measure.label(
            image
        )

        return labeled_image

    def _get_quantized_channels(
        self,
        channels
    ):
        quantized_channels = {}

        for channel_name, channel_array in channels.items():
            quantized_channels[channel_name] = self._quantize_channel(
                channel_array
            )
        return quantized_channels

    def _get_connected_components_(
        self,
        channels
    ):
        connected_components = {}

        for channel_name, channel_array in channels.items():

            connected_components[
                channel_name] = self._label_components(
                channels[channel_name]
            )
        return connected_components

    def _get_connected_blobs_skimage(
        self,
        connected_components,
        list_of_index,
        key
    ):
        blobs = connected_components[key]
        # for i, j in enumerate(blobs):
        #    blob[i] = self._image[j]

        connected_regions = {}
        for x in list_of_index[key]:
            sl = ndimage.find_objects(blobs == x)
            region_selected = self._image[sl[0]]
            connected_regions[x] = region_selected
        return connected_regions

    def _is_reasonable_area(
        self,
        region
    ):
        return self._region_low_thresh < region.area < self._region_high_thresh

    def _is_reasonable_intensity(
        self,
        region
    ):
        return region.mean_intensity > self._intensity_threshold

    def _area_and_intensity(
        self,
        region
    ):
        return all(
            [
                self._is_reasonable_area(region)  # ,
                # self._is_reasonable_intensity(region),
            ]
        )

    # def _filter_regions(
    #    self,
    #    labeled_image,
    #    quantized_image
    #    criteria
    # ):
    #    region_df = pd.DataFrame(
    #        skimage.measure.regionprops_table(
    #       labeled_image,
    #       quantized_image,
    #       properties=['area', 'mean_intensity']
    #    )
    #    )

    def _filter_list_index(
        self,
        labeled_image,
        image,
    ):
        region_df = pd.DataFrame(
            measure.regionprops_table(
                labeled_image,
                image,
                properties=['label', 'area', 'mean_intensity']
            )
        )
        filtered_areas_df = region_df[
            (region_df["area"] >= self._region_low_thresh) &
            (region_df["area"] <= self._region_high_thresh)
        ]

        return filtered_areas_df["label"].tolist()

    def _filter_connected_components(
        self,
        connected_components,
        criteria
    ):
        filtered_cc = {}
        for key, labeled_image in connected_components.items():
            filtered_cc[key] = self._filter_list_index(
                labeled_image,
                self._image
            )
        return filtered_cc

    def __call__(
        self,
        image=None,
        bins=6,
        blur=True,
        area_low_ratio=.02,
        area_high_ratio=.4,
        intensity_threshold=150
    ):

        self._bins = bins
        self._image = load_image(image)
        self._image_area = self._image.shape[0] * self._image.shape[1]
        if blur:
            self._image = cv2.blur(self._image, (5, 5))

        self._region_low_thresh = self._image_area*area_low_ratio
        self._region_high_thresh = self._image_area*area_high_ratio
        self._intensity_threshold = intensity_threshold

        self._image_rgb = cv2.cvtColor(
            self._image,
            cv2.COLOR_BGR2RGB
        )

        self._image_hsv = cv2.cvtColor(
            self._image_rgb,
            cv2.COLOR_BGR2HSV
        )

        quantized_image = self._quantize_image(self._image_hsv)

        (h_quantized_image,
         s_quantized_image,
         v_quantized_image) = cv2.split(quantized_image)

        quantized_image_channels = {
            "h_quantized_image": h_quantized_image,
            "s_quantized_image": s_quantized_image,
            "v_quantized_image": v_quantized_image
        }

        connected_components_quantized_image = self._get_connected_components_(
            quantized_image_channels
        )

        quantized_image_indices = self._filter_connected_components(
            connected_components_quantized_image,
            criteria=self._area_and_intensity
        )

        h, s, v = cv2.split(self._image_hsv)

        quantized_h = quantize_image(h)
        quantized_s = quantize_image(s)
        quantized_v = quantize_image(v)

        quantized_channels = {
            "h_quantized": quantized_h,
            "s_quantized": quantized_s,
            "v_quantized": quantized_v
        }

        connected_components_quantized_channels = self._get_connected_components_(
            quantized_channels
        )

        quantized_channels_indices = self._filter_connected_components(
            connected_components_quantized_channels,
            criteria=self._area_and_intensity
        )

        if self._debug:
            images = {
                 "hsv": self._image_hsv,
                # "hsv_quantized": quantized_image["hsv"],
                "h": h,
                "s": s,
                "v": v,
                "h_quantized_image": h_quantized_image,
                "s_quantized_image": s_quantized_image,
                "v_quantized_image": v_quantized_image,
                "h_quantized": quantized_h,
                "s_quantized": quantized_s,
                "v_quantized": quantized_v,
            }

            plot_figures(images)
            plt.show()
            #key = "s_quantized"

            connected_regions = self._get_connected_blobs_skimage(
                connected_components_quantized_image,
                quantized_image_indices,
                key="h_quantized_image"
            )
            print(connected_regions)

            plot_figures(connected_regions, 5, 5)
            plt.show()

            connected_regions = self._get_connected_blobs_skimage(
                connected_components_quantized_channels,
                quantized_channels_indices,
                key="h_quantized"
            )
            print(connected_regions)

            plot_figures(connected_regions, 2, 2)
            plt.show()

        return None



import matplotlib.pyplot as plt
import numpy as np
import cv2
from skimage import measure
from sabhi_utils.image_utils.utils import load_image


class BlurDetector:

    def __init__(
        self,
        preprocessor=None,
        thresh=20
    ):

        self._preprocessor = preprocessor
        self._thresh = thresh

    def _get_image_center(self):

        return (int(self._w / 2.0), int(self._h / 2.0))

    def _visualize(self):
        # check to see if we are visualizing our output
        # compute the magnitude spectrum of the transform
        # display the original input image
        (fig, ax) = plt.subplots(1, 2, )
        ax[0].imshow(self._image, cmap="gray")
        ax[0].set_title("Input")
        ax[0].set_xticks([])
        ax[0].set_yticks([])
        # display the magnitude image
        ax[1].imshow(self._magnitude, cmap="gray")
        ax[1].set_title("Magnitude Spectrum")
        ax[1].set_xticks([])
        ax[1].set_yticks([])
        # show our plots
        plt.show()

    def _zero_center_inverse(self):
        fft_shift = self._fft_shift.copy()

        fft_shift[
            self._cY - self._size:self._cY + self._size,
            self._cX - self._size:self._cX + self._size
        ] = 0

        fft_shift = np.fft.ifftshift(fft_shift)
        recon = np.fft.ifft2(fft_shift)
        return recon

    def __call__(
        self,
        image,
        vis=False
    ):

        image = load_image(image)
        self._image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        self._h, self._w = self._image.shape
        self._cX, self._cY = self._get_image_center()

        self._size = int(self._h/8)

        self._fft = np.fft.fft2(self._image)
        self._fft_shift = np.fft.fftshift(self._fft)

        self._recon = self._zero_center_inverse()

        self._magnitude = 20 * np.log(
            np.abs(self._recon)
        )

        self._mean = np.mean(self._magnitude)

        if vis:
            self._visualize()

        # the image will be considered "blurry" if the mean value of the
        # magnitudes is less than the threshold value
        # return (self._mean <= self._thresh, self._mean)
        return {
            'isDetected': self._mean <= self._thresh
        }


class BrightspotDetector:
    def __init__(
        self,
        preprocess=[]
    ):

        self._preprocess = preprocess

    def __call__(
        self,
        image,
        percentage_limit=0.02,
        debug=False
    ):

        image = load_image(image)
        self._image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        self._blurred = cv2.GaussianBlur(self._image, (11, 11), 0)
        self._size = self._image.shape[0] * self._image.shape[1]
        self._percentage_limit = percentage_limit
        self._blob_size_limit = int(self._size * self._percentage_limit)

        # threshold the image to reveal light regions in the
        # blurred image

        self._thresh = cv2.threshold(
            self._blurred, 238, 255, cv2.THRESH_BINARY)[1]

        self._thresh = cv2.erode(self._thresh, None, iterations=2)
        self._thresh = cv2.dilate(self._thresh, None, iterations=4)

        # perform a connected component analysis on the thresholded
        # image, then initialize a mask to store only the "large"
        # components
        labels = measure.label(self._thresh, background=0)
        if debug:
            cv2.imshow("self._thresh", self._thresh)
            cv2.waitKey(0)
            print(labels)
        #mask = np.zeros(self._thresh.shape, dtype="uint8")
        # loop over the unique components
        for label in np.unique(labels):
            # if this is the background label, ignore it
            if label == 0:
                continue
            # otherwise, construct the label mask and count the
            # number of pixels
            labelMask = np.zeros(self._thresh.shape, dtype="uint8")
            labelMask[labels == label] = 255
            numPixels = cv2.countNonZero(labelMask)
            # if the number of pixels in the component is sufficiently
            # large, then add it to our mask of "large blobs"
            if numPixels > self._blob_size_limit:
                return {
                    'isDetected': True
                }

        # return (False, "Insignificant bright spots")
        return {
            'isDetected': False
        }



from PIL import Image
import cv2
from sabhi_utils.utility import initial_logger
from sabhi_utils.utility import check_and_read_gif, get_image_file_list
import imutils
import numpy as np
import base64
from math import sqrt, ceil, floor

from sklearn.cluster import KMeans

import numpy as np
import cv2

from typing import Union
from typing import Tuple

import urllib.request
import base64

import operator

logger = initial_logger()


def get_image(image_path, resize_width=None):

    img, flag = check_and_read_gif(image_path)
    if not flag:
        img = cv2.imread(image_path)

        if resize_width:
            img = imutils.resize(img, width=resize_width)
        return img
    if img is None:
        logger.info("error in loading image:{}".format(image_path))


def get_hsv_means_for_windows(image, N_h=5, N_w=10, debug=False):

    h_means = []
    s_means = []
    image_slices = slice_image(image, row=N_w, col=N_h)
    for slice in image_slices:
        slice_cv = cv2.cvtColor(np.array(slice), cv2.COLOR_RGB2HSV)
        if debug:
            cv2.imshow("slice_cv", slice_cv)
            cv2.waitKey(0)
        h, s, _ = cv2.split(slice_cv)
        h_means.append(h.mean())
        s_means.append(s.mean())

    return np.column_stack((h_means, s_means))


def calc_columns_rows(n):
    """
    Calculate the number of columns and rows required to divide an image
    into ``n`` parts.
    Return a tuple of integers in the format (num_columns, num_rows)
    """
    num_columns = int(ceil(sqrt(n)))
    num_rows = int(ceil(n / float(num_columns)))
    return (num_columns, num_rows)


def validate_image(image, number_tiles):
    """Basic sanity checks prior to performing a split."""
    TILE_LIMIT = 99 * 99

    try:
        number_tiles = int(number_tiles)
    except BaseException:
        raise ValueError("number_tiles could not be cast to integer.")

    if number_tiles > TILE_LIMIT or number_tiles < 2:
        raise ValueError(
            "Number of tiles must be between 2 and {} (you \
                          asked for {}).".format(
                TILE_LIMIT, number_tiles
            )
        )


def validate_image_col_row(image, col, row):
    """Basic checks for columns and rows values"""
    SPLIT_LIMIT = 99

    try:
        col = int(col)
        row = int(row)
    except BaseException:
        raise ValueError(
            "columns and rows values could not be cast to integer.")

    if col < 1 or row < 1 or col > SPLIT_LIMIT or row > SPLIT_LIMIT:
        raise ValueError(
            "Number of columns and rows must be between 1 and"
            #f"{SPLIT_LIMIT} (you asked for rows: {row} and col: {col})."
        )
    if col == 1 and row == 1:
        raise ValueError(
            "There is nothing to divide. You asked for the entire image.")


def slice_image(
    image,
    number_tiles=None,
    col=None,
    row=None,
    DecompressionBombWarning=True,
):
    """
    Split an image into a specified number of tiles.
    Args:
       image (cv2 array):  The image to split.
       number_tiles (int):  The number of tiles required.
    Kwargs:
       save (bool): Whether or not to save tiles to disk.
       DecompressionBombWarning (bool): Whether to suppress
       Pillow DecompressionBombWarning
    Returns:
        Tuple of :class:`Tile` instances.
    """
    if DecompressionBombWarning is False:
        Image.MAX_IMAGE_PIXELS = None

    im = Image.fromarray(image)
    im_w, im_h = im.size

    columns = 0
    rows = 0
    if number_tiles:
        validate_image(im, number_tiles)
        columns, rows = calc_columns_rows(number_tiles)
    else:
        validate_image_col_row(im, col, row)
        columns = col
        rows = row

    tile_w, tile_h = int(floor(im_w / columns)), int(floor(im_h / rows))

    tiles = []
    number = 1
    for pos_y in range(0, im_h - rows, tile_h):  # -rows for rounding error.
        for pos_x in range(0, im_w - columns, tile_w):  # as above.
            area = (pos_x, pos_y, pos_x + tile_w, pos_y + tile_h)
            image = im.crop(area)
            #position = (int(floor(pos_x / tile_w)) + 1, int(floor(pos_y / tile_h)) + 1)
            #coords = (pos_x, pos_y)
            tiles.append(image)
            number += 1
    return tuple(tiles)


def is_grayscale(image):
    return len(image.shape) == 2


def pad(
    image,
    height=500,
    width=500,
    color=(0, 0, 0)
):
    # read image
    hh, ww = height, width

    ht, wd, cc = image.shape

    result = np.full((hh, ww, cc), color, dtype=np.uint8)

    # compute center offset
    xx = (ww - wd) // 2
    yy = (hh - ht) // 2

    # copy img image into center of result image
    result[yy:yy+ht, xx:xx+wd] = image

    return result


def pad_unet(
    image: np.array,
    factor: int = 32,
    border: int = cv2.BORDER_REFLECT_101
) -> tuple:
    """Pads the image on the sides, so that it will be divisible by factor.
    Common use case: UNet type architectures.

    Args:
        image:
        factor:
        border: cv2 type border.

    Returns: padded_image

    """
    height, width = image.shape[:2]

    if height % factor == 0:
        y_min_pad = 0
        y_max_pad = 0
    else:
        y_pad = factor - height % factor
        y_min_pad = y_pad // 2
        y_max_pad = y_pad - y_min_pad

    if width % factor == 0:
        x_min_pad = 0
        x_max_pad = 0
    else:
        x_pad = factor - width % factor
        x_min_pad = x_pad // 2
        x_max_pad = x_pad - x_min_pad

    padded_image = cv2.copyMakeBorder(
        image, y_min_pad, y_max_pad, x_min_pad, x_max_pad, border)

    return padded_image, (x_min_pad, y_min_pad, x_max_pad, y_max_pad)


def unpad_unet(image: np.array, pads: Tuple[int, int, int, int]) -> np.ndarray:
    """Crops patch from the center so that sides are equal to pads.

    Args:
        image:
        pads: (x_min_pad, y_min_pad, x_max_pad, y_max_pad)

    Returns: cropped image

    """
    x_min_pad, y_min_pad, x_max_pad, y_max_pad = pads
    height, width = image.shape[:2]

    return image[y_min_pad: height - y_max_pad, x_min_pad: width - x_max_pad]


def isBase64(sb):
    try:
        if isinstance(sb, str):
            # If there's any unicode here, an exception will be thrown and the function will return false
            sb_bytes = bytes(sb, 'ascii')
        elif isinstance(sb, bytes):
            sb_bytes = sb
        else:
            raise ValueError("Argument must be string or bytes")
        return base64.b64encode(base64.b64decode(sb_bytes)) == sb_bytes
    except Exception:
        return False


def load_image(
    image,
    resize_width=None
):

    if isBase64(image):
        image = bytes(image, 'ascii')
        im_bytes = base64.b64decode(image)
        nparr = np.frombuffer(im_bytes, dtype=np.uint8)
        result = cv2.imdecode(nparr, cv2.IMREAD_COLOR)

    elif isinstance(image, str):
        result = get_image(
            image,
            resize_width=resize_width
        )

    elif type(image) == np.ndarray:
        result = image

    else:
        return None

    return result


def crop_image(
    image=None,
    y=None,
    x=None,
    h=None,
    w=None
):

    image = load_image(image)
    cropped_image = image[y:y+h, x:x+w]

    return cropped_image


def get_bounding_box_points(pts):
    """
    Function for getting the bounding box points in the correct
    order

    Params
    pts     The points in the bounding box. (x, y) coordinates

    Returns
    rect    The ordered set of points
    """
    # initialzie a list of coordinates that will be ordered such that
    # 1st point -> Top left
    # 2nd point -> Top right
    # 3rd point -> Bottom right
    # 4th point -> Bottom left
    rect = np.zeros((4, 2), dtype="float32")

    # the top-left point will have the smallest sum, whereas
    # the bottom-right point will have the largest sum
    s = pts.sum(axis=1)
    rect[0] = pts[np.argmin(s)]
    rect[2] = pts[np.argmax(s)]
# now, compute the difference between the points, the
    # top-right point will have the smallest difference,
    # whereas the bottom-left will have the largest difference
    diff = np.diff(pts, axis=1)
    rect[1] = pts[np.argmin(diff)]
    rect[3] = pts[np.argmax(diff)]

    # return the ordered coordinates
    return rect


def warp_birds_eye(
    image=None,
    rect=None,
    output_process=False
):
    (tl, tr, br, bl) = rect
    # compute the width of the new image, which will be the
    # maximum distance between bottom-right and bottom-left
    # x-coordiates or the top-right and top-left x-coordinates
    widthA = np.sqrt(
        ((br[0] - bl[0]) ** 2) + ((br[1] - bl[1]) ** 2)
    )
    widthB = np.sqrt(
        ((tr[0] - tl[0]) ** 2) + ((tr[1] - tl[1]) ** 2)
    )
    maxWidth = max(int(widthA), int(widthB))

    # compute the height of the new image, which will be the
    # maximum distance between the top-right and bottom-right
    # y-coordinates or the top-left and bottom-left y-coordinates
    heightA = np.sqrt(((tr[0] - br[0]) ** 2) + ((tr[1] - br[1]) ** 2))
    heightB = np.sqrt(((tl[0] - bl[0]) ** 2) + ((tl[1] - bl[1]) ** 2))
    maxHeight = max(int(heightA), int(heightB))

    # now that we have the dimensions of the new image, construct
    # the set of destination points to obtain a "birds eye view",
    # (i.e. top-down view) of the image, again specifying points
    # in the top-left, top-right, bottom-right, and bottom-left
    # order
    dst = np.array(
        [
            [0, 0],  # Top left point
            [maxWidth - 1, 0],  # Top right point
            [maxWidth - 1, maxHeight - 1],  # Bottom right point
            [0, maxHeight - 1],
        ],  # Bottom left point
        dtype="float32",  # Date type
    )

    # compute the perspective transform matrix and then apply it
    M = cv2.getPerspectiveTransform(rect, dst)
    warped = cv2.warpPerspective(image, M, (maxWidth, maxHeight))

    if output_process:
        cv2.imwrite("output/deskewed.jpg", warped)

    # return the warped image
    return warped


def quantize_image(
    image,
    clusters=6
):

    if len(image.shape) > 2:
        arr = image.reshape((-1, 3))
    else:
        arr = image
    kmeans = KMeans(n_clusters=clusters).fit(arr)
    labels = kmeans.labels_
    centers = kmeans.cluster_centers_
    less_colors = centers[labels].reshape(image.shape).astype('uint8')

    return less_colors


def get_image_area(
    image
):
    return image.shape[0] * image.shape[1]


def np_to_base64(
    np_array
):
    _, im_arr = cv2.imencode('.jpg', np_array)
    im_bytes = im_arr.tobytes()
    im_b64 = base64.b64encode(im_bytes)
    return im_b64.decode("utf-8")


def load_rgb(
    image: np.array,
    lib: str = "cv2"
) -> np.array:
    """Load RGB image from path.

    Args:
        image_path: path to image
        lib: library used to read an image.
            currently supported `cv2` and `jpeg4py`

    Returns: 3 channel array with RGB image

    """
    image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
    return image


def resize(image, width=None, height=None, inter=cv2.INTER_AREA):
    # initialize the dimensions of the image to be resized and # grab the image size
    dim = None
    (h, w) = image.shape[:2]

    # if both the width and height are None, then return the
    # original image
    if width is None and height is None:
        return image

    # check to see if the width is None
    if width is None:
        # calculate the ratio of the height and construct the
        # dimensions
        r = height / float(h)
        dim = (int(w * r), height)

    # otherwise, the height is None
    else:
        # calculate the ratio of the width and construct the
        # dimensions
        r = width / float(w)
        dim = (width, int(h * r))

    # resize the image
    resized = cv2.resize(image, dim, interpolation=inter)

    # return the resized image
    return resized


def image_url_to_cv2_image(url):
    # download the image, convert it to a NumPy array, and then read
    # it into OpenCV format

    resp = image_url_to_byte(url)
    image = np.asarray(
        bytearray(resp),
        dtype="uint8"
    )
    image = cv2.imdecode(image, cv2.IMREAD_COLOR)

    return image


def np_to_base64(
    np_array
):
    _, im_arr = cv2.imencode('.jpg', np_array)
    im_bytes = im_arr.tobytes()
    im_b64 = base64.b64encode(im_bytes)
    return im_b64.decode("utf-8")


def get_aspect_ratio_check(
    image=None,
    height=None,
    width=None,
    aspect_ratio=16/10,
    threshold=.05,
    check=operator.lt
):

    response = False

    h, w = height, width
    if image is not None:
        h, w = image.shape[:2]

    try:
        ratio = w/h

        abs_ratio = abs(1-ratio/aspect_ratio),

        response = check(
            abs(1-ratio/aspect_ratio),
            threshold
        )
    except ZeroDivisionError:
        response = False

    return response


def get_image_ratio(
    image_a,
    image_b,
):
    r = None
    image_a_height, image_a_width = image_a.shape[:2]

    image_b_height, image_b_width = image_b.shape[:2]

    r = image_a_height/image_b_height

    return r


def scaled_point(
    point,
    r
):

    return np.round(point * r).astype(int)


def get_scaled_points(
    intersections=None,
    orig_image=None,
    processed_image=None
):
    ratio = get_image_ratio(
        orig_image,
        processed_image
    )

    scaled_intersection = intersections * ratio

    return np.round(scaled_intersection)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup, find_packages
from io import open

with open('requirements.txt', encoding="utf-8-sig") as f:
    requirements = f.readlines()


def readme():
    with open("README.md", encoding="utf-8-sig") as f:
        README = f.read()
    return README


setup(
    name='sabhi_utils',
    packages=find_packages(),
    include_package_data=True,
    version='2.0.3',
    install_requires=requirements,
    license='Apache License 2.0',
    description='Self rolled utils to be used with the Sabhi ML Services',
    long_description=readme(),
    long_description_content_type='text/markdown',
    url='',
    download_url='',
    keywords=[
        'image utils text processing'
    ],
    classifiers=[
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Utilities'
    ], )



import cv2
import math
from scipy import ndimage
import numpy as np

import time 
#import pytesseract
#import re
#import imutils

class RotationCorrector:
    def __init__(self, output_process=False):
        self.output_process = output_process

    def __call__(self, image):
        img_before = image.copy()

        img_edges = cv2.Canny(img_before, 50, 150, apertureSize=3)

        lines = cv2.HoughLinesP(
            img_edges, 1, math.pi / 90.0, 100, minLineLength=100, maxLineGap=5
        )
        print("Number of lines found:", len(lines))

        def get_angle(line):
            x1, y1, x2, y2 = line[0]
            return math.degrees(math.atan2(y2 - y1, x2 - x1))

        median_angle = np.median(
            np.array([get_angle(line) for line in lines])
        )
        img_rotated = ndimage.rotate(
            img_before, median_angle, cval=255, reshape=False)

        print("Angle is {}".format(median_angle))

        if self.output_process:
            cv2.imwrite("output/10. tab_extract rotated.jpg", img_rotated)

        return img_rotated


class Resizer:
    """Resizes image.

    Params
    ------
    image   is the image to be resized
    height  is the height the resized image should have.
            Width is changed by similar ratio.

    Returns
    -------
    Resized image
    """

    def __init__(self, height=1280, output_process=False):
        self._height = height
        self.output_process = output_process

    def __call__(self, image):
        #        if image.shape[0] <= self._height:
        #            return image
        ratio = round(self._height / image.shape[0], 3)
        width = int(image.shape[1] * ratio)
        dim = (width, self._height)
        resized = cv2.resize(image, dim, interpolation=cv2.INTER_AREA)
        if self.output_process:
            cv2.imwrite("output/resized.jpg", resized)
        return resized


class OtsuThresholder:
    """Thresholds image by using the otsu method

    Params
    ------
    image   is the image to be Thresholded

    Returns
    -------
    Thresholded image
    """

    def __init__(self, thresh1=0, thresh2=255, output_process=False):
        self.output_process = output_process
        self.thresh1 = thresh1
        self.thresh2 = thresh2

    def __call__(self, image):
        image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        _, thresholded = cv2.threshold(
            image,
            self.thresh1,
            self.thresh2,
            cv2.THRESH_BINARY + cv2.THRESH_OTSU
        )
        if self.output_process:
            cv2.imwrite("output/thresholded.jpg", thresholded)
        return thresholded


class AdaptiveThresholder:
    """Thresholds image by using the Gaussian  method

    Params
    ------
    image   is the image to be Thresholded

    Returns
    -------
    Thresholded image
    """

    def __init__(self, thresh1=255, block_size=11, constant=2, output_process=False):
        self.output_process = output_process
        self.thresh1 = thresh1
        self.block_size = block_size
        self.constant = constant

    def __call__(self, image):
        image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        thresholded = cv2.adaptiveThreshold(
            image,
            self.thresh1,
            cv2.ADAPTIVE_THRESH_GAUSSIAN_C,
            cv2.THRESH_BINARY,
            self.block_size,
            self.constant
        )

        if self.output_process:
            cv2.imwrite("output/thresholded.jpg", thresholded)
        return thresholded


class FastDenoiser:
    """Denoises image by using the fastNlMeansDenoising method

    Params
    ------
    image       is the image to be Thresholded
    strength    the amount of denoising to apply

    Returns
    -------
    Denoised image
    """

    def __init__(self, strength=11, output_process=False):
        self._strength = strength
        self.output_process = output_process

    def __call__(self, image):
        temp = cv2.fastNlMeansDenoising(image, h=self._strength)
        if self.output_process:
            cv2.imwrite("output/denoised.jpg", temp)
        return temp


class Closer:
    def __init__(self, kernel_size=3, iterations=10, output_process=False):
        self._kernel_size = kernel_size
        self._iterations = iterations
        self.output_process = output_process

    def __call__(self, image):
        kernel = cv2.getStructuringElement(
            cv2.MORPH_ELLIPSE, (self._kernel_size, self._kernel_size)
        )
        closed = cv2.morphologyEx(
            image, cv2.MORPH_CLOSE, kernel, iterations=self._iterations
        )

        if self.output_process:
            cv2.imwrite("output/closed.jpg", closed)
        return closed


class Opener:
    def __init__(self, kernel_size=3, iterations=25, output_process=False):
        self._kernel_size = kernel_size
        self._iterations = iterations
        self.output_process = output_process

    def __call__(self, image):
        kernel = cv2.getStructuringElement(
            cv2.MORPH_ELLIPSE, (self._kernel_size, self._kernel_size)
        )
        opened = cv2.morphologyEx(
            image, cv2.MORPH_OPEN, kernel, iterations=self._iterations
        )

        if self.output_process:
            cv2.imwrite("output/opened.jpg", opened)
        return opened


class EdgeDetector:
    def __init__(self, output_process=False):
        self.output_process = output_process

    def __call__(self, image, thresh1=50, thresh2=150, apertureSize=3):
        edges = cv2.Canny(image, thresh1, thresh2, apertureSize=apertureSize)
        if self.output_process:
            cv2.imwrite("/home/hamza/code/SabhiUtils/output/"+str(time.time())+"edges.jpg", edges)
        return edges


class SharpenAndDilate:
    def __init__(
        self,
        kernel=(7, 7),
        normalized=False,
        output_process=False
    ):
        self._kernel = kernel
        self._normalized = normalized
        self.output_process = output_process

    def __call__(
        self,
        image,
        debug=False
    ):

        self._image = image
        self._debug = debug

        return self._sharpened_and_dialated()

    def _sharpened_and_dialated(self):
        rgb_planes = cv2.split(self._image)
        result = []
        for plane in rgb_planes:
            image_result = cv2.dilate(plane,
                                      np.ones(self._kernel, np.uint8))
            bg_img = cv2.medianBlur(image_result, 21)

            if self._debug:
                cv2.imshow("bg_img", bg_img)
                cv2.waitKey(0)
            image_result = 255 - cv2.absdiff(plane, bg_img)

            if self._normalized:
                norm_img = cv2.normalize(
                    image_result,
                    None,
                    alpha=0,
                    beta=255,
                    norm_type=cv2.NORM_MINMAX,
                    dtype=cv2.CV_8UC1)

                result.append(norm_img)
            else:
                result.append(image_result)
        image = cv2.merge(result)
        return image
